import aiosqlite

from assistants.config.environment import DB_TABLE
from assistants.user_data.sqlite_backend.assistants import \
    TABLE_NAME as ASSISTANTS
from assistants.user_data.sqlite_backend.chat_history import \
    TABLE_NAME as CHAT_HISTORY


async def init_db():
    async with aiosqlite.connect(DB_TABLE) as db:
        await db.execute(
            f"CREATE TABLE IF NOT EXISTS {CHAT_HISTORY} (chat_id INTEGER PRIMARY KEY, history TEXT);"
        )
        await db.execute(
            f"CREATE TABLE IF NOT EXISTS {ASSISTANTS} (assistant_name TEXT PRIMARY KEY, assistant_id TEXT, config_hash TEXT);"
        )
        await db.commit()


if __name__ == "__main__":
    import asyncio

    asyncio.run(init_db())



IO_INSTRUCTIONS = """\
-h, --help:   Show this help message
-e:           Open the default editor to compose a prompt
-c:           Copy the previous response to the clipboard
-cb:          Copy the code blocks from the previous response to the clipboard
-n:           Start a new thread and clear the terminal screen
clear:        (or CTRL+L) Clear the terminal screen without starting a new thread"""



import asyncio

from assistants.telegram_ui.sqlite_user_data import SqliteUserData
from assistants.user_data.sqlite_backend import init_db

if __name__ == "__main__":
    asyncio.run(init_db())
    user_data = SqliteUserData()
    asyncio.run(user_data.create_db())



import sys

from assistants.cli import output
from assistants.cli.constants import IO_INSTRUCTIONS
from assistants.lib.exceptions import ConfigError

try:
    from assistants.config import environment
except ConfigError as e:
    import re

    pattern = re.compile(r"Missing required (\w+) environment variable")

    match = pattern.match(str(e))

    if match is None:
        output.fail(f"Error: {e}")
        sys.exit(1)

    output.fail(
        f"`{match.group(1)}` not found! Check README.md for setup instructions. Exiting..."
    )
    sys.exit(1)

import asyncio
import re
import sys
from pathlib import Path
from typing import Optional, Union, cast

import pyperclip
from openai.types.beta.threads import Message
from prompt_toolkit import prompt
from prompt_toolkit.history import FileHistory
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.styles import Style

from assistants import version
from assistants.ai.assistant import Assistant, Completion
from assistants.cli.arg_parser import get_args
from assistants.cli.terminal import clear_screen
from assistants.cli.utils import (
    PERSISTENT_THREAD_ID_FILE,
    get_text_from_default_editor,
    get_thread_id,
    highlight_code_blocks,
)
from assistants.lib.exceptions import NoResponseError

bindings = KeyBindings()
history = FileHistory(f"{Path.home()}/.ai-assistant-history")
style = Style.from_dict(
    {"": "ansigreen", "input": "ansibrightgreen"},
)
message = [("class:input", ">>> ")]


@bindings.add("c-l")
def _(event):
    clear_screen()


def _io_loop(
    assistant: Union[Assistant, Completion],
    initial_input: str = "",
    last_message: Optional[Message] = None,
    thread_id: Optional[str] = None,
):
    user_input = ""
    is_completion = isinstance(assistant, Completion)

    def get_user_input() -> str:
        global message
        return prompt(message, style=style, history=history)  # type: ignore

    while initial_input or (user_input := get_user_input()).lower() not in {
        "q",
        "quit",
        "exit",
    }:
        output.reset()
        if initial_input:
            output.user_input(initial_input)
            user_input = initial_input
            initial_input = ""

        if not user_input.strip():
            continue

        else:
            user_input = user_input.strip()

        match user_input.lower().strip():
            case instruction if instruction in {"-h", "--help", "help"}:
                output.inform(IO_INSTRUCTIONS)
                continue
            case "-e":
                user_input = get_text_from_default_editor().strip()
                if not user_input:
                    continue
                output.user_input(user_input)
            case "-c":
                if is_completion:
                    previous_response = assistant.memory[-1]["content"]  # type: ignore

                elif not last_message:
                    previous_response = ""
                else:
                    previous_response = last_message.content[0].text.value  # type: ignore

                if not previous_response:
                    output.warn("No previous message to copy.")
                    continue

                try:
                    pyperclip.copy(previous_response)
                except pyperclip.PyperclipException:
                    output.fail(
                        "Error copying to clipboard; this feature doesn't seem to be "
                        "available in the current terminal environment."
                    )
                    continue

                output.inform("Copied response to clipboard")
                continue
            case "-cb":
                if is_completion:
                    previous_response = assistant.memory[-1]["content"]  # type: ignore

                elif not last_message:
                    previous_response = ""
                else:
                    previous_response = last_message.content[0].text.value  # type: ignore

                if not previous_response:
                    output.warn("No previous message to copy from.")
                    continue

                code_blocks = re.split(
                    r"(```.*?```)", previous_response, flags=re.DOTALL
                )
                code_only = [
                    "\n".join(block.split("\n")[1:-1]).strip()
                    for block in code_blocks
                    if block.startswith("```")
                ]

                if not code_only:
                    output.warn("No codeblocks in previous message!")
                    continue

                all_code = "\n\n".join(code_only)

                try:
                    pyperclip.copy(all_code)
                except pyperclip.PyperclipException:
                    output.fail(
                        "Error copying code to clipboard; this feature doesn't seem to "
                        "be available in the current terminal environment."
                    )
                    continue

                output.inform("Copied code blocks to clipboard")
                continue
            case "-n":
                thread_id = None
                last_message = None
                clear_screen()
                continue
            case "clear":
                clear_screen()
                continue
            case nothing if not nothing:
                continue

        message = asyncio.run(
            assistant.converse(
                initial_input or user_input,
                last_message.thread_id if last_message else thread_id,
            )
        )

        if message is None:
            output.warn("No response from the AI model.")
            continue

        if is_completion:
            output.default(message.content)  # type: ignore
            output.new_line(2)
            continue

        message = cast(Message, message)
        if last_message and message and message.id == last_message.id:
            raise NoResponseError

        text = highlight_code_blocks(message.content[0].text.value)

        output.default(text)
        output.new_line(2)
        last_message = message

        if last_message and not thread_id:
            with open(PERSISTENT_THREAD_ID_FILE, "w") as file:
                file.write(last_message.thread_id)


def cli():
    last_message = None
    thread_id = None

    args = get_args()

    instructions = (
        args.instructions if args.instructions else environment.ASSISTANT_INSTRUCTIONS
    )
    initial_input = " ".join(args.prompt) if args.prompt else None

    if args.continue_thread:
        thread_id = get_thread_id()
        if thread_id is None:
            output.warn("Warning: could not read last thread id; starting new thread.")
    output.default(
        f"AI Assistant v{version.__VERSION__}; type 'help' for a list of commands.\n"
    )
    if args.editor:
        # Open the default editor to compose formatted prompt
        initial_input = get_text_from_default_editor(initial_input)
    elif args.input_file:
        # Read the initial prompt from a file
        try:
            with open(args.f, "r") as file:
                initial_input = file.read()
        except FileNotFoundError:
            output.fail(f"Error: The file '{args.f}' was not found.")
            sys.exit(1)

    # Create the assistant
    if args.code:
        assistant = Completion(model=environment.CODE_MODEL)
    else:
        assistant = Assistant(
            name="AI Assistant",
            model=environment.DEFAULT_MODEL,
            instructions=instructions,
            tools=[{"type": "code_interpreter"}],
        )

    # IO Loop
    try:
        _io_loop(assistant, initial_input, last_message, thread_id)
    except (EOFError, KeyboardInterrupt):
        # Exit gracefully if ctrl+C or ctrl+D are pressed
        sys.exit(0)



import os

from assistants.lib.exceptions import ConfigError
from assistants.log import logger

ASSISTANTS_API_KEY_NAME = os.environ.get("ASSISTANTS_API_KEY_NAME", "OPENAI_API_KEY")

try:
    OPENAI_API_KEY = os.environ[ASSISTANTS_API_KEY_NAME]
    if not OPENAI_API_KEY:
        raise KeyError
except KeyError as e:
    error = f"Missing required {ASSISTANTS_API_KEY_NAME} environment variable"
    logger.error(error)
    raise ConfigError(error) from e

DEFAULT_MODEL = os.environ.get("DEFAULT_MODEL", "gpt-4o-mini")
CODE_MODEL = os.environ.get("CODE_MODEL", "o1-mini")
IMAGE_MODEL = os.environ.get("IMAGE_MODEL", "dall-e-3")
ASSISTANT_INSTRUCTIONS = os.environ.get(
    "ASSISTANT_INSTRUCTIONS", "You are a helpful assistant."
)
HOME_DIR = os.environ["HOME"]
DB_TABLE = os.environ.get("USER_DATA_DB", f"{HOME_DIR}/.assistants_user_data.db")



import os
import re
import subprocess
import tempfile

import pygments
from pygments import highlight
from pygments.formatters import TerminalFormatter
from pygments.lexers import get_lexer_by_name

PERSISTENT_THREAD_ID_FILE = f"{os.environ.get('HOME', '')}/.assistant-last-thread-id"


def highlight_code_blocks(markdown_text):
    code_block_pattern = re.compile(r"```(\w+)?\n(.*?)```", re.DOTALL)

    def replacer(match):
        lang = match.group(1)
        code = match.group(2)
        if lang:
            lexer = get_lexer_by_name(lang, stripall=True)
        else:
            lexer = get_lexer_by_name("text", stripall=True)
        return f"```{lang if lang else ''}\n{highlight(code, lexer, TerminalFormatter())}```"

    return code_block_pattern.sub(replacer, markdown_text)


def get_thread_id():
    try:
        with open(PERSISTENT_THREAD_ID_FILE, "r") as file:
            return file.read().strip()
    except FileNotFoundError:
        return None


def get_text_from_default_editor(initial_text=None):
    # Create a temporary file
    with tempfile.NamedTemporaryFile(suffix=".md", delete=False) as temp_file:
        temp_file_path = temp_file.name

    if initial_text:
        with open(temp_file_path, "w") as text_file:
            text_file.write(initial_text)

    # Open the editor for the user to input text
    editor = os.environ.get("EDITOR", "nano")
    subprocess.run([editor, temp_file_path])

    # Read the contents of the file after the editor is closed
    with open(temp_file_path, "r") as file:
        text = file.read()

    # Remove the temporary file
    os.remove(temp_file_path)

    return text



import os
import sys

from assistants.log import logger

try:
    from assistants.telegram_ui import build_bot, run_polling
except ImportError:
    logger.error(
        "Could not import required modules. Install with `pip install assistants[telegram]`"
    )
    sys.exit(1)


BOT_TOKEN = os.getenv("TG_BOT_TOKEN")


def main():
    if BOT_TOKEN is None:
        print("Please set the TG_BOT_TOKEN environment variable.")
        return

    run_polling(build_bot(BOT_TOKEN))


if __name__ == "__main__":
    main()



import os
from functools import wraps

import aiohttp
from telegram import Update
from telegram.constants import ParseMode
from telegram.ext import (Application, ApplicationBuilder, CommandHandler,
                          ContextTypes, MessageHandler, filters)

from assistants.ai.assistant import Assistant
from assistants.config.environment import (ASSISTANT_INSTRUCTIONS,
                                           DEFAULT_MODEL, OPENAI_API_KEY)
from assistants.telegram_ui.sqlite_user_data import SqliteUserData
from assistants.telegram_ui.user_data import ChatHistory, NotAuthorized

user_data = SqliteUserData()

assistant = Assistant(
    name=os.getenv("ASSISTANT_NAME"),
    model=DEFAULT_MODEL,
    instructions=ASSISTANT_INSTRUCTIONS,
    tools=[{"type": "code_interpreter"}],
    api_key=OPENAI_API_KEY,
)


def restricted_access(f):
    @wraps(f)
    async def wrapper(update: Update, *args, **kwargs):
        try:
            await user_data.check_chat_authorised(update.effective_chat.id)
        except NotAuthorized:
            await user_data.check_user_authorised(update.effective_user.id)

        return await f(update, *args, **kwargs)

    return wrapper


def requires_superuser(f):
    @wraps(f)
    async def wrapper(update: Update, *args, **kwargs):
        await user_data.check_superuser(update.effective_user.id)

        return await f(update, *args, **kwargs)

    return wrapper


def requires_reply_to_message(f):
    @wraps(f)
    async def wrapper(update: Update, context: ContextTypes.DEFAULT_TYPE):
        try:
            return await f(update, context)
        except AttributeError:
            await context.bot.send_message(
                chat_id=update.effective_chat.id,
                text="You must reply to a message from the target user to use this command",
            )

    return wrapper


@requires_superuser
@requires_reply_to_message
async def promote_user(update: Update, context: ContextTypes.DEFAULT_TYPE):
    await user_data.promote_superuser(update.message.reply_to_message.from_user.id)
    await context.bot.send_message(
        chat_id=update.effective_chat.id, text="User promoted"
    )


@requires_superuser
@requires_reply_to_message
async def demote_user(update: Update, context: ContextTypes.DEFAULT_TYPE):
    await user_data.demote_superuser(update.message.reply_to_message.from_user.id)
    await context.bot.send_message(
        chat_id=update.effective_chat.id, text="User demoted"
    )


@requires_superuser
async def authorise_chat(update: Update, context: ContextTypes.DEFAULT_TYPE):
    await user_data.authorise_chat(update.effective_chat.id)
    await context.bot.send_message(
        chat_id=update.effective_chat.id, text="Chat authorised"
    )


@requires_superuser
@requires_reply_to_message
async def authorise_user(update: Update, context: ContextTypes.DEFAULT_TYPE):
    await user_data.authorise_chat(update.message.reply_to_message.from_user.id)
    await context.bot.send_message(
        chat_id=update.effective_chat.id, text="User authorised"
    )


@requires_superuser
async def deauthorise_chat(update: Update, context: ContextTypes.DEFAULT_TYPE):
    await user_data.deauthorise_chat(update.effective_chat.id)
    await context.bot.send_message(
        chat_id=update.effective_chat.id, text="Chat de-authorised"
    )


@requires_superuser
@requires_reply_to_message
async def deauthorise_user(update: Update, context: ContextTypes.DEFAULT_TYPE):
    await user_data.deauthorise_user(update.message.reply_to_message.from_user.id)
    await context.bot.send_message(
        chat_id=update.effective_chat.id, text="User de-authorised"
    )


@restricted_access
async def new_thread(update: Update, context: ContextTypes.DEFAULT_TYPE):
    await user_data.clear_last_thread_id(update.effective_chat.id)
    assistant.last_message_id = None
    await context.bot.send_message(
        update.effective_chat.id, "Conversation history cleared."
    )


@restricted_access
async def toggle_auto_reply(update: Update, context: ContextTypes.DEFAULT_TYPE):
    chat_history = await user_data.get_chat_history(update.effective_chat.id)
    result = "OFF" if chat_history.auto_reply else "ON"
    await user_data.set_auto_reply(
        update.effective_chat.id, not chat_history.auto_reply
    )
    await context.bot.send_message(
        chat_id=update.effective_chat.id, text=f"Auto reply is {result}"
    )


@restricted_access
async def message_handler(update: Update, context: ContextTypes.DEFAULT_TYPE):
    existing_chat = await user_data.get_chat_history(update.effective_chat.id)
    message_text = update.message.text
    if not existing_chat.auto_reply:
        bot_id = context.bot.id
        bot_username = f"@{context.bot.username}"
        if bot_username not in message_text and (
            not update.message.reply_to_message
            or update.message.reply_to_message.from_user.id != bot_id
        ):
            return
        message_text = message_text.replace(
            bot_username, os.getenv("ASSISTANT_NAME", "[ASSISTANT NAME]")
        )

    response_message = await assistant.converse(message_text, existing_chat.thread_id)

    if not existing_chat.thread_id:
        await user_data.save_chat_history(
            ChatHistory(
                chat_id=update.effective_chat.id,
                thread_id=response_message.thread_id,
                auto_reply=existing_chat.auto_reply,
            )
        )

    response = response_message.content[0].text.value

    response_parts = response.split("```")

    if len(response_parts) % 2 == 0:
        # Should be an odd number of parts if codeblocks
        await context.bot.send_message(
            chat_id=update.effective_chat.id,
            text=response,
        )
        return

    for i, part in enumerate(response_parts):
        if i % 2:
            await context.bot.send_message(
                chat_id=update.effective_chat.id,
                text=f"```{part}```",
                parse_mode=ParseMode.MARKDOWN_V2,
            )
        else:
            await context.bot.send_message(
                chat_id=update.effective_chat.id,
                text=part,
            )


@restricted_access
async def generate_image(update: Update, context: ContextTypes.DEFAULT_TYPE):
    prompt = update.message.text.replace("/image ", "")
    image_url = await assistant.image_prompt(prompt)

    async with aiohttp.ClientSession() as session:
        async with session.get(image_url) as response:
            image_content = await response.read()

    await update.message.reply_photo(image_content)


def build_bot(token: str) -> Application:
    application = ApplicationBuilder().token(token).build()
    application.add_handler(CommandHandler("add_chat", authorise_chat))
    application.add_handler(CommandHandler("remove_chat", deauthorise_chat))
    application.add_handler(CommandHandler("add_user", authorise_user))
    application.add_handler(CommandHandler("remove_user", deauthorise_user))
    application.add_handler(CommandHandler("promote", promote_user))
    application.add_handler(CommandHandler("demote", demote_user))
    application.add_handler(CommandHandler("new_thread", new_thread))
    application.add_handler(CommandHandler("auto_reply", toggle_auto_reply))
    application.add_handler(CommandHandler("image", generate_image))
    application.add_handler(MessageHandler(filters.TEXT, message_handler))
    return application


def run_polling(application: Application):
    application.run_polling()


if __name__ == "__main__":
    BOT_TOKEN = os.getenv("TG_BOT_TOKEN")
    run_polling(build_bot(BOT_TOKEN))



"""
Assistant classes encapsulating interactions with the OpenAI API(s)
"""
import asyncio
import hashlib
from typing import Optional, TypedDict

import openai
from openai._types import NOT_GIVEN, NotGiven
from openai.types.beta import Thread
from openai.types.beta.threads import Message, Run
from openai.types.chat import ChatCompletionMessage

from assistants.config import environment
from assistants.lib.exceptions import NoResponseError
from assistants.log import logger
from assistants.user_data.sqlite_backend.assistants import (get_assistant_data,
                                                            save_assistant_id)


class Assistant:  # pylint: disable=too-many-instance-attributes
    """
    Encapsulates interactions with the OpenAI Assistants API.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        *,
        name: str,
        model: str,
        instructions: str,
        tools: list | NotGiven = NOT_GIVEN,
        api_key: str = environment.OPENAI_API_KEY,
    ):
        """
        Initialize the Assistant instance.

        :param name: The name of the assistant.
        :param model: The model to be used by the assistant.
        :param instructions: Instructions for the assistant.
        :param tools: Optional tools for the assistant.
        :param api_key: API key for OpenAI.
        """
        self.client = openai.OpenAI(
            api_key=api_key, default_headers={"OpenAI-Beta": "assistants=v2"}
        )
        self.instructions = instructions
        self.tools = tools
        self.model = model
        self.name = name
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
        self._config_hash = None
        self.assistant = loop.run_until_complete(self.load_or_create_assistant())
        self.last_message_id = None

    @property
    def config_hash(self):
        """
        A hash of the current config options to prevent regeneration of the same assistant.

        :return: The configuration hash.
        """
        if not self._config_hash:
            self._config_hash = self._generate_config_hash()
        return self._config_hash

    def _generate_config_hash(self):
        """
        Generate a hash based on the current configuration.

        :return: The generated hash.
        """
        return hashlib.sha256(
            f"{self.instructions}{self.model}{self.tools}".encode()
        ).hexdigest()

    async def load_or_create_assistant(self):
        """
        Get any existing assistant ID / config hash from the database or create a new assistant.

        :return: The assistant object.
        """
        existing_id, config_hash = await get_assistant_data(self.name, self.config_hash)
        if existing_id:
            try:
                assistant = self.client.beta.assistants.retrieve(existing_id)
                if config_hash != self.config_hash:
                    logger.info("Config has changed, updating assistant...")
                    self.client.beta.assistants.update(
                        existing_id,
                        instructions=self.instructions,
                        model=self.model,
                        tools=self.tools,
                        name=self.name,
                    )
                    await save_assistant_id(self.name, assistant.id, self.config_hash)
                return assistant
            except openai.NotFoundError:
                pass
        assistant = self.client.beta.assistants.create(
            name=self.name,
            instructions=self.instructions,
            model=self.model,
            tools=self.tools,
        )
        await save_assistant_id(self.name, assistant.id, self.config_hash)
        return assistant

    def _create_thread(self, messages=NOT_GIVEN) -> Thread:
        """
        Create a new thread.

        :param messages: Optional initial messages for the thread.
        :return: The created thread.
        """
        return self.client.beta.threads.create(messages=messages)

    def _new_thread(self) -> Thread:
        """
        Create a new thread.

        :return: The created thread.
        """
        thread = self._create_thread()
        return thread

    def start_thread(self, prompt: str) -> Thread:
        """
        Create a new thread and add the first message to it.

        :param prompt: The initial prompt for the thread.
        :return: The created thread.
        """
        thread = self._new_thread()
        self.client.beta.threads.messages.create(
            thread_id=thread.id,
            role="user",
            content=prompt,
        )
        return thread

    def continue_thread(self, prompt: str, thread_id: str) -> Message:
        """
        Add a new message to an existing thread.

        :param prompt: The message content.
        :param thread_id: The ID of the thread to continue.
        :return: The created message.
        """
        message = self.client.beta.threads.messages.create(
            thread_id=thread_id,
            role="user",
            content=prompt,
        )
        return message

    def run_thread(self, thread: Thread) -> Run:
        """
        Run the thread using the current assistant.

        :param thread: The thread to run.
        :return: The run object.
        """
        run = self.client.beta.threads.runs.create(
            thread_id=thread.id,
            assistant_id=self.assistant.id,
        )
        return run

    def _get_thread(self, thread_id: str) -> Thread:
        """
        Retrieve a thread by its ID.

        :param thread_id: The ID of the thread to retrieve.
        :return: The retrieved thread.
        """
        return self.client.beta.threads.retrieve(thread_id)

    async def prompt(self, prompt: str, thread_id: Optional[str] = None) -> Run:
        """
        Create a message using the prompt and add it to an existing thread or create a new thread.

        :param prompt: The message content.
        :param thread_id: Optional ID of the thread to continue.
        :return: The run object.
        """
        if thread_id is None:
            thread = self.start_thread(prompt)
            run = self.run_thread(thread)
            thread_id = thread.id
        else:
            thread = self._get_thread(thread_id)
            self.continue_thread(prompt, thread_id)
            run = self.run_thread(thread)
        while run.status in {"queued", "in_progress"}:
            run = self.client.beta.threads.runs.retrieve(
                thread_id=thread_id,
                run_id=run.id,
            )
            await asyncio.sleep(0.5)
        return run

    async def image_prompt(self, prompt: str) -> Optional[str]:
        """
        Request an image to be generated using a separate image model.

        :param prompt: The image prompt.
        :return: The URL of the generated image.
        """
        response = self.client.images.generate(
            model=environment.IMAGE_MODEL,
            prompt=prompt,
            size="1024x1024",
            quality="standard",
            n=1,
        )
        return response.data[0].url

    async def converse(
        self, user_input: str, thread_id: Optional[str] = None
    ) -> Optional[Message]:
        """
        Converse with the assistant by creating or continuing a thread.

        :param user_input: The user's input message.
        :param thread_id: Optional ID of the thread to continue.
        :return: The last message in the thread.
        """
        if not user_input:
            return None

        if thread_id is None:
            run = await self.prompt(user_input)
            thread_id = run.thread_id
        else:
            await self.prompt(user_input, thread_id)

        messages = self.client.beta.threads.messages.list(
            thread_id=thread_id, order="asc", after=self.last_message_id or NOT_GIVEN
        ).data

        last_message_in_thread = messages[-1]

        if last_message_in_thread.id == self.last_message_id:
            raise NoResponseError

        self.last_message_id = last_message_in_thread.id

        return last_message_in_thread


class MessageDict(TypedDict):
    """
    Typed dictionary for message data.
    """

    role: str
    content: str | None


class Completion:
    """
    Encapsulates interactions with the OpenAI Chat Completion API.
    """

    def __init__(
        self,
        model: str,
        max_memory: int = 50,
        api_key: str = environment.OPENAI_API_KEY,
    ):
        """
        Initialize the Completion instance.

        :param model: The model to be used for completions.
        :param max_memory: Maximum number of messages to retain in memory.
        :param api_key: API key for OpenAI.
        """
        self.client = openai.OpenAI(api_key=api_key)
        self.model = model
        self.memory: list[MessageDict] = []
        self.max_memory = max_memory

    def truncate_memory(self):
        """
        Truncate the memory to the maximum allowed messages.
        """
        self.memory = self.memory[-self.max_memory :]

    def complete(self, prompt: str) -> ChatCompletionMessage:
        """
        Generate a completion for the given prompt.

        :param prompt: The prompt to complete.
        :return: The completion message.
        """
        new_prompt = MessageDict(
            role="user",
            content=prompt,
        )
        self.truncate_memory()
        self.memory.append(new_prompt)
        response = self.client.chat.completions.create(
            model=self.model,
            messages=self.memory,  # type: ignore
        )
        message = response.choices[0].message
        self.memory.append({"role": "assistant", "content": message.content})
        return response.choices[0].message

    async def converse(
        self, user_input: str, *args, **kwargs  # pylint: disable=unused-argument
    ) -> ChatCompletionMessage:
        """
        Converse with the assistant using the chat completion API.

        :param user_input: The user's input message.
        :return: The completion message.
        """
        return self.complete(user_input)



import argparse

from assistants import version


def get_args():
    parser = argparse.ArgumentParser(
        description=f"CLI for AI Assistant v{version.__VERSION__}"
    )
    parser.add_argument(
        "-e",
        "--editor",
        action="store_true",
        help="Open the default editor to compose a prompt.",
    )
    parser.add_argument(
        "-f",
        "--input-file",
        metavar="INPUT_FILE",
        type=str,
        help="Read the initial prompt from a file (e.g., 'input.txt').",
    )
    parser.add_argument(
        "-i",
        "--instructions",
        metavar="INSTRUCTIONS_FILE",
        type=str,
        help="Read the initial instructions (system message) from a specified file; "
        "if not provided, environment variable `ASSISTANT_INSTRUCTIONS` or defaults "
        "will be used.",
    )
    parser.add_argument(
        "-t",
        "--continue-thread",
        action="store_true",
        help="Continue previous thread. (not currently possible with `-C` option)",
    )
    parser.add_argument(
        "-C",
        "--code",
        action="store_true",
        help="Use specialised reasoning/code model. WARNING: This model will be slower "
        "and more expensive to use.",
    )
    parser.add_argument(
        "prompt",
        nargs="*",
        help="Positional arguments concatenate into a single prompt. E.g. `ai-cli "
        "Is this a single prompt\\?` (question mark escaped)\n"
        "...will be passed to the program as a single string (without the backslash). You "
        "can also use quotes to pass a single argument with spaces and special characters. "
        "See the -e and -f options for more advanced prompt options.",
    )
    args = parser.parse_args()
    return args



import os
from abc import ABC, abstractmethod
from typing import Optional

from pydantic import BaseModel

DB = os.environ.get("TG_USER_DATA")


class NotAuthorized(ValueError):
    pass


class ChatHistory(BaseModel):
    chat_id: int
    thread_id: Optional[str] = None
    auto_reply: bool


class UserData(ABC):
    DB = DB

    @abstractmethod
    async def create_db(self):
        pass

    @abstractmethod
    async def get_chat_history(self, chat_id: int) -> ChatHistory:
        pass

    @abstractmethod
    async def save_chat_history(self, history: ChatHistory):
        pass

    @abstractmethod
    async def check_user_authorised(self, user_id: int):
        pass

    @abstractmethod
    async def check_superuser(self, user_id: int):
        pass

    @abstractmethod
    async def authorise_user(self, user_id: int):
        pass

    @abstractmethod
    async def promote_superuser(self, user_id: int):
        pass

    @abstractmethod
    async def demote_superuser(self, user_id: int):
        pass

    @abstractmethod
    async def authorise_chat(self, chat_id: int):
        pass

    @abstractmethod
    async def deauthorise_user(self, user_id: int):
        pass

    @abstractmethod
    async def deauthorise_chat(self, chat_id: int):
        pass

    @abstractmethod
    async def clear_last_thread_id(self, chat_id: int):
        pass
